/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-builtin.c,v 1.28 2009/05/28 01:37:30 zlb Exp $ */

/* The built-in solver.
 * Currently it only supports solving diagonal matrices for phgDofCopy() */

#include "phg.h"

#define Initialize		NULL
#define Finalize		NULL
#define Init			NULL
#define RegisterOptions		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

static int
Create(SOLVER *solver)
{
    assert(solver->mat->type != PHG_MATRIX_FREE);
    return 0;
}

static int
Destroy(SOLVER *solver)
{
    phgMatDestroy(&solver->mat);
    solver->sent_to_oem = TRUE;

    return 0;
}

static int
Assemble(SOLVER *solver)
{
    return 0;
}
 
static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    MAT *mat = solver->mat;
    MAP *map = mat->rmap;
    MAT_ROW *row;
    FLOAT *rhs, *p;
    INT i, n0;

    if (mat->type == PHG_PACKED)
	phgMatUnpack(mat);
    assert(mat->type == PHG_UNPACKED);

    if (map->nlocal <= 0)
	return 0;

    p = x->data;
    n0 = map->partition[map->rank];
    row = mat->rows;
    rhs = solver->rhs->data;
    for (i = 0; i < map->nlocal; i++, row++, rhs++, p++) {
	if (row->ncols == 0)
	    phgError(1, "empty row, global index = %d\n", i + n0);
	if (row->ncols != 1 || *row->cols != i)
	    phgError(1, "the built-in solver only solves diagonal systems.\n");
	if (*row->data == 0.0)
	    phgError(1, "%s:%d, zero diagonal entry found in the matrix.\n",
			__FILE__, __LINE__);
	*p = *rhs / *row->data;
    }

    if (destroy) {
	phgMatDestroy(&solver->mat);
	solver->sent_to_oem = TRUE;
    }

    solver->residual = 0.0;	/* fake residual norm */
    return solver->nits = 1;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverBuiltIn_ = {
    "Built-in", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy, AddMatrixEntries,
    AddRHSEntries, Assemble, Solve, NULL, NULL, NULL, FALSE, TRUE
};
